import warnings
from abc import ABC, abstractmethod
from functools import partial
from typing import Any

from numpy.typing import NDArray
from scipy.sparse import csc_matrix, csr_matrix


class AbstractSolver(ABC):
    @abstractmethod
    def factor(self, m: csc_matrix | csr_matrix) -> None:
        ...

    @abstractmethod
    def solve(self, rhs: NDArray, transpose: bool = False) -> NDArray:
        ...

    @abstractmethod
    def clear(self) -> None:
        ...


class SciPySolver(AbstractSolver):
    _ctx: dict[Any] = {}

    def __init__(self, **options):
        from scipy.sparse.linalg import splu

        self._ctx["splu"] = partial(splu, **options)

    def factor(self, m: csc_matrix) -> None:
        self._ctx["factorization"] = self._ctx["splu"](m)

    def solve(self, rhs: NDArray, transpose: bool = False) -> NDArray:
        return self._ctx["factorization"].solve(rhs, trans="T" if transpose else "N")

    def clear(self) -> None:
        self._ctx.pop("factorization", None)


class PardisoSolver(AbstractSolver):
    _ctx: dict[Any] = {}

    def __init__(self, **options):
        from pyMKL import pardisoSolver

        self._ctx["pardiso"] = partial(pardisoSolver, **options)

    def factor(self, m: csr_matrix | csc_matrix) -> None:
        self._ctx["factorization"] = self._ctx["pardiso"](m)
        self._ctx["factorization"].factor()

    def solve(self, rhs: NDArray, transpose: bool = False) -> NDArray:
        """Warning! Transpose does nothing because pyMKL does not expose iparm
        This is not a problem in the case of tofea since matrices are guaranteed to be
        symmetric, but gradients will be wrong if that is not the case.
        Probably will need to implement a wrapper for MKL at some point.
        """
        return self._ctx["factorization"].solve(rhs)

    def clear(self) -> None:
        if "factorization" in self._ctx:
            self._ctx["factorization"].clear()
        self._ctx.pop("factorization", None)


class CholeskySolver(AbstractSolver):
    _ctx: dict[Any] = {}

    def __init__(self, **options):
        from sksparse.cholmod import cholesky

        self._ctx["cholesky"] = partial(cholesky, **options)

    def factor(self, m: csc_matrix) -> None:
        if "factorization" in self._ctx:
            self._ctx["factorization"].cholesky_inplace(m)
        else:
            self._ctx["factorization"] = self._ctx["cholesky"](m)

    def solve(self, rhs: NDArray, transpose: bool = False) -> NDArray:
        """Transpose does nothing because Cholesky decomposition only works for real
        symmetric matrices...
        """
        return self._ctx["factorization"].solve_A(rhs)

    def clear(self) -> None:
        self._ctx.pop("factorization", None)


class UmfpackSolver(AbstractSolver):
    _ctx: dict[Any] = {}

    def __init__(self, **options):
        from scikits.umfpack import UMFPACK_A, UMFPACK_At, UmfpackContext

        self._ctx["modes"] = {"N": UMFPACK_A, "T": UMFPACK_At}
        self._ctx["umfpack"] = UmfpackContext(**options)

    def factor(self, m: csc_matrix) -> None:
        self._ctx["umfpack"].numeric(m)

    def solve(self, rhs: NDArray, transpose: bool = False) -> NDArray:
        mode = self._ctx["modes"]["T"] if transpose else self._ctx["modes"]["N"]
        return self._ctx["umfpack"].solve(
            mode, self._ctx["umfpack"].mtx, rhs, autoTranspose=True
        )

    def clear(self) -> None:
        self._ctx["umfpack"].free_numeric()


class GPUSolver(AbstractSolver):
    _ctx: dict[Any] = {}

    def __init__(self, **options):
        import cupy as cp
        from cupyx.scipy.sparse import csr_matrix
        from cupyx.scipy.sparse.linalg import cg

        self._ctx["cg"] = partial(cg, **options)
        self._ctx["ascp"] = cp.array
        self._ctx["asnp"] = cp.asnumpy
        self._ctx["ascps"] = csr_matrix

    def factor(self, m: csc_matrix) -> None:
        self._ctx["m"] = self._ctx["ascps"](m)

    def solve(self, rhs: NDArray, transpose: bool = False) -> NDArray:
        m = self._ctx["m"].T if transpose else self._ctx["m"]
        b = self._ctx["ascp"](rhs)
        x = self._ctx["cg"](m, b)[0]
        return self._ctx["asnp"](x)

    def clear(self) -> None:
        self._ctx.pop("m", None)


scipy_solver = SciPySolver(
    diag_pivot_thresh=0.1,
    permc_spec="MMD_AT_PLUS_A",
    options=dict(SymmetricMode=True),
)

try:
    pardiso_solver = PardisoSolver(mtype=11)
except ImportError:
    warnings.warn("pyMKL not found, pardiso_solver unavailable")

try:
    cholesky_solver = CholeskySolver()
except ImportError:
    warnings.warn("scikit-sparse not found, cholesky_solver unavailable")

try:
    umfpack_solver = UmfpackSolver()
except ImportError:
    warnings.warn("scikit-umfpack not found, umfpack_solver unavailable")

try:
    gpu_solver = GPUSolver(tol=1e-12)
except ImportError:
    warnings.warn("cupy not found, gpu_solver unavailable")


def get_solver(solver):
    match solver:
        case "scipy":
            return scipy_solver
        case "pardiso":
            return pardiso_solver
        case "cholesky":
            return cholesky_solver
        case "umfpack":
            return umfpack_solver
        case "gpu":
            return gpu_solver
        case _:
            raise ValueError(f"Invalid solver: {solver}")



from dataclasses import dataclass
from functools import cached_property
from typing import Iterable

import numpy as np
import sympy
from numpy.typing import NDArray

__all__ = ["Q4Element_K", "Q4Element_T", "H8Element_K", "H8Element_T"]


@dataclass(frozen=True, slots=True, kw_only=True)
class Element:
    dx: float = 0.5
    dy: float = 0.5
    dz: float = 0.5
    eps: float = 1e-6
    dtype: type = np.float64

    @staticmethod
    def _b_entries(
        rule: Iterable[int],
        shape_funcs: Iterable[sympy.Expr],
        clist: Iterable[sympy.Symbol],
    ) -> tuple[sympy.Expr]:
        shape_list = np.concatenate([x * np.asarray(rule) for x in shape_funcs])
        return tuple(map(sympy.diff, shape_list, clist))


class Q4Element(Element):
    @property
    def symbols(self) -> tuple[sympy.Symbol, ...]:
        return sympy.symbols("a b x y", real=True)

    @property
    def shape_funcs(self) -> list[sympy.Expr]:
        a, b, x, y = self.symbols
        return [
            (a - x) * (b - y) / (4 * a * b),
            (a + x) * (b - y) / (4 * a * b),
            (a + x) * (b + y) / (4 * a * b),
            (a - x) * (b + y) / (4 * a * b),
        ]


class H8Element(Element):
    @property
    def symbols(self) -> tuple[sympy.Symbol, ...]:
        return sympy.symbols("a b c x y z", real=True)

    @property
    def shape_funcs(self) -> list[sympy.Expr]:
        a, b, c, x, y, z = self.symbols
        return [
            (a - x) * (b - y) * (c - z) / (8 * a * b * c),
            (a + x) * (b - y) * (c - z) / (8 * a * b * c),
            (a + x) * (b + y) * (c - z) / (8 * a * b * c),
            (a - x) * (b + y) * (c - z) / (8 * a * b * c),
            (a - x) * (b - y) * (c + z) / (8 * a * b * c),
            (a + x) * (b - y) * (c + z) / (8 * a * b * c),
            (a + x) * (b + y) * (c + z) / (8 * a * b * c),
            (a - x) * (b + y) * (c + z) / (8 * a * b * c),
        ]


@dataclass(frozen=True, slots=True)
class Q4Element_K(Q4Element):
    e: float = 1.0
    nu: float = 1 / 3

    @cached_property
    def element(self) -> NDArray:
        a, b, x, y = self.symbols
        E, nu = sympy.symbols("E nu", real=True)

        B = sympy.Matrix(
            [
                self._b_entries([1, 0], self.shape_funcs, 8 * [x]),
                self._b_entries([0, 1], self.shape_funcs, 8 * [y]),
                self._b_entries([1, 1], self.shape_funcs, 4 * [y, x]),
            ]
        )

        C = (E / (1 - nu**2)) * sympy.Matrix(
            [[1, nu, 0], [nu, 1, 0], [0, 0, (1 - nu) / 2]]
        )

        dK = B.T * C * B
        K = dK.integrate((x, -a, a), (y, -b, b))
        K = np.array(
            K.subs({a: self.dx, b: self.dy, E: self.e, nu: self.nu}),
            dtype=self.dtype,
        )
        K[np.abs(K) < self.eps] = 0
        return K


@dataclass(frozen=True, slots=True)
class Q4Element_T(Q4Element):
    k: float = 1.0

    @cached_property
    def element(self) -> NDArray:
        a, b, x, y = self.symbols
        k = sympy.symbols("k", real=True)

        B = sympy.Matrix(
            [
                self._b_entries([1], self.shape_funcs, 4 * [x]),
                self._b_entries([1], self.shape_funcs, 4 * [y]),
            ]
        )

        C = sympy.Matrix([[k, 0], [0, k]])

        dK = B.T * C * B
        K = dK.integrate((x, -a, a), (y, -b, b))
        K = np.array(K.subs({a: self.dx, b: self.dy, k: self.k}), dtype=self.dtype)
        K[np.abs(K) < self.eps] = 0
        return K


@dataclass(frozen=True, slots=True)
class H8Element_K(H8Element):
    e: float = 1.0
    nu: float = 1 / 3

    @property
    def G(self) -> float:
        return self.e / (2 * (1 + self.nu))

    @property
    def g(self) -> float:
        return self.e / ((1 + self.nu) * (1 - 2 * self.nu))

    @cached_property
    def element(self) -> NDArray:
        a, b, c, x, y, z = self.symbols
        E, nu, g, G = sympy.symbols("E nu g G", real=True)
        o = sympy.symbols("o", real=True)  # dummy symbol

        B = sympy.Matrix(
            [
                self._b_entries([1, 0, 0], self.shape_funcs, 24 * [x]),
                self._b_entries([0, 1, 0], self.shape_funcs, 24 * [y]),
                self._b_entries([0, 0, 1], self.shape_funcs, 24 * [z]),
                self._b_entries([1, 1, 1], self.shape_funcs, 8 * [y, x, o]),
                self._b_entries([1, 1, 1], self.shape_funcs, 8 * [o, z, y]),
                self._b_entries([1, 1, 1], self.shape_funcs, 8 * [z, o, x]),
            ]
        )

        C = sympy.Matrix(
            [
                [(1 - nu) * g, nu * g, nu * g, 0, 0, 0],
                [nu * g, (1 - nu) * g, nu * g, 0, 0, 0],
                [nu * g, nu * g, (1 - nu) * g, 0, 0, 0],
                [0, 0, 0, G, 0, 0],
                [0, 0, 0, 0, G, 0],
                [0, 0, 0, 0, 0, G],
            ]
        )

        dK = B.T * C * B
        K = dK.integrate((x, -a, a), (y, -b, b), (z, -c, c))
        K = np.array(
            K.subs(
                {
                    a: self.dx,
                    b: self.dy,
                    c: self.dz,
                    E: self.e,
                    nu: self.nu,
                    g: self.g,
                    G: self.G,
                }
            ),
            dtype=self.dtype,
        )
        K[np.abs(K) < self.eps] = 0
        return K


@dataclass(frozen=True, slots=True)
class H8Element_T(H8Element):
    k: float = 1.0

    @cached_property
    def element(self) -> NDArray:
        a, b, c, x, y, z = self.symbols
        k = sympy.symbols("k")

        B = sympy.Matrix(
            [
                self._b_entries([1], self.shape_funcs, 8 * [x]),
                self._b_entries([1], self.shape_funcs, 8 * [y]),
                self._b_entries([1], self.shape_funcs, 8 * [z]),
            ]
        )

        C = sympy.Matrix([[k, 0, 0], [0, k, 0], [0, 0, k]])

        dK = B.T * C * B
        K = dK.integrate((x, -a, a), (y, -b, b), (z, -c, c))
        K = np.array(
            K.subs({a: self.dx, b: self.dy, c: self.dz, k: self.k}),
            dtype=self.dtype,
        )
        K[np.abs(K) < self.eps] = 0
        return K



from functools import cached_property
from pathlib import Path
from time import perf_counter

import autograd.numpy as anp
import cupy as cp
import cupyx.scipy.sparse as cs
import cupyx.scipy.sparse.linalg as cl
import h5py
import numpy as np
from autograd.extend import defvjp, primitive
from numpy.typing import NDArray
from pyMKL import pardisoSolver
from scipy.sparse import coo_matrix
from scipy.sparse.linalg import cg

from tofea.primitives import solve_coo

from .elements import H8Element_K, H8Element_T


class FEA3D_K:
    def __init__(
        self,
        domain,
        dofs,
        fixed,
        load,
        z_chunks=10,
        solver="direct",
        element="./data/H8_K.h5",
    ):
        self.solver = solver

        self.shape = domain
        self.forces = load.ravel()
        self.fixdofs = dofs[fixed].ravel()
        self.freedofs = dofs[~fixed].ravel()
        self.element_stiffness = self.load_element(element)
        self.displacement = np.zeros(dofs.size)
        self.chunks = z_chunks

        defvjp(self.compliance, self.compliance_vjp)

    @staticmethod
    def load_element(fp):
        fp = Path(fp)
        if fp.is_file():
            with h5py.File(fp, "r") as f:
                K = np.array(f["K"])
        else:
            fp.parent.mkdir(exist_ok=True, parents=True)
            K = H8Element_K().element
            with h5py.File(fp, "w") as f:
                f.create_dataset("K", data=K, compression="gzip")
        return K

    @cached_property
    def dofmap(self):
        nelx, nely, nelz = self.shape
        a = np.arange(3) + 3
        b = np.arange(3 * (nely + 2), 3 * (nely + 2) + 3)
        c = np.arange(3 * (nely + 1), 3 * (nely + 1) + 3)
        d = np.arange(3)
        e = np.arange(
            3 * ((nelx + 1) * (nely + 1) + 1),
            3 * ((nelx + 1) * (nely + 1) + 1) + 3,
        )
        f = np.arange(
            3 * ((nelx + 1) * (nely + 1) + (nely + 2)),
            3 * ((nelx + 1) * (nely + 1) + (nely + 2)) + 3,
        )
        g = np.arange(
            3 * ((nelx + 1) * (nely + 1) + (nely + 1)),
            3 * ((nelx + 1) * (nely + 1) + (nely + 1)) + 3,
        )
        h = np.arange(3 * (nelx + 1) * (nely + 1), 3 * (nelx + 1) * (nely + 1) + 3)
        return np.r_[a, b, c, d, e, f, g, h]

    @staticmethod
    def inverse_permutation(indices):
        inverse_perm = np.zeros(len(indices), dtype=np.int64)
        inverse_perm[indices] = np.arange(len(indices), dtype=np.int64)
        return inverse_perm

    def _get_dof_indices(self, k_ylist, k_xlist):
        index_map = self.inverse_permutation(
            np.concatenate([self.freedofs, self.fixdofs])
        )
        keep = np.isin(k_xlist, self.freedofs) & np.isin(k_ylist, self.freedofs)
        i = index_map[k_ylist][keep]
        j = index_map[k_xlist][keep]
        return index_map, keep, np.stack([i, j])

    def global_stiffness(self, x):
        i, j = np.meshgrid(range(len(self.dofmap)), range(len(self.dofmap)))
        nelx, nely, nelz = x.shape

        mat = None
        for chunk in np.array_split(np.arange(nelz), self.chunks)[::-1]:
            vals, ix, iy = [], [], []
            for elz in chunk:
                for elx in range(nelx):
                    for ely in range(nely):
                        e2sdofmap = self.dofmap + x.ndim * (
                            ely + elx * (nely + 1) + elz * (nelx + 1) * (nely + 1)
                        )
                        v = x[elx, ely, elz] * self.element_stiffness
                        ix.append(e2sdofmap[i])
                        iy.append(e2sdofmap[j])
                        vals.append(v[i, j])
            vals = np.asarray(vals).ravel()
            ix = np.asarray(ix, dtype=int).ravel()
            iy = np.asarray(iy, dtype=int).ravel()

            _, keep, indices = self._get_dof_indices(ix, iy)
            if mat is None:
                mat = coo_matrix((vals[keep], indices)).tocsr()
            else:
                mat += coo_matrix((vals[keep], indices), shape=mat.shape).tocsr()
        return mat

    @staticmethod
    @primitive
    def compliance(x, self):
        nelx, nely, nelz = x.shape
        X, Y, Z = np.indices(x.shape)
        X *= nely + 1
        Z *= (nelx + 1) * (nely + 1)
        e2sdofmap = np.expand_dims(self.dofmap.reshape(-1, 1, 1), axis=1)
        e2sdofmap = np.add(e2sdofmap, x.ndim * (X + Y + Z))

        print("assemble matrix...", end=" ")
        t0 = time()
        Kfree = self.global_stiffness(x)
        print(f"done: {time() - t0:.3f}s")

        if self.solver == "direct":
            print("direct solve...", end=" ")
            t0 = time()
            solver = pardisoSolver(Kfree, mtype=2)
            self.displacement[self.freedofs] = solver.run_pardiso(
                phase=13, rhs=self.forces[self.freedofs]
            )
            solver.clear()
            print(f"done: {time() - t0:.3f}s")
        elif self.solver == "iterative":
            print("iterative solve...", end=" ")
            t0 = time()
            self.displacement[self.freedofs], _ = cg(Kfree, self.forces[self.freedofs])
            print(f"done: {time() - t0:.3f}s")
        elif self.solver == "gpu":
            print("gpu solve...", end=" ")
            t0 = time()
            self.displacement[self.freedofs] = cp.asnumpy(
                cl.cg(cs.csr_matrix(Kfree), cp.array(self.forces[self.freedofs]))[0]
            )
            print(f"done: {time() - t0:.3f}s")
        else:
            raise RuntimeError(f"Unknown solver: {self.solver}")

        Qe = self.displacement[e2sdofmap]
        QeK = np.tensordot(Qe, self.element_stiffness, axes=(0, 0))
        Qe_T = np.swapaxes(Qe.T, 2, 0)
        self._QeKQe = np.einsum("klmn,klmn->klm", QeK, Qe_T)
        return np.sum(x * self._QeKQe)

    @staticmethod
    def compliance_vjp(ans, x, self):
        del ans, x

        def vjp(g):
            return -g * self._QeKQe

        return vjp

    def __call__(self, x):
        return self.compliance(x, self)


class FEA3D_T:
    dof_dim = 1

    def __init__(self, fixed, solver="scipy"):
        nz, nx, ny = fixed.shape[:3]
        dofs = np.arange(fixed.size, dtype=np.uint32).reshape(fixed.shape)
        self.out_shape = (nx - 1, ny - 1, nz - 1)
        self.fixdofs = dofs[fixed].ravel()
        self.freedofs = dofs[~fixed].ravel()
        self.solver = solver

    @cached_property
    def element(fp):
        return H8Element_T().element

    @cached_property
    def dofmap(self):
        nelx, nely, _ = self.out_shape
        e2s = np.r_[1, (nely + 2), (nely + 1), 0]
        return np.r_[e2s, (e2s + (nelx + 1) * (nely + 1))]

    @cached_property
    def keep_indices(
        self,
    ) -> tuple[NDArray[np.bool_], NDArray[np.uint32]]:
        i, j = np.meshgrid(range(len(self.dofmap)), range(len(self.dofmap)))
        ix = self.e2sdofmap[:, i].ravel()
        iy = self.e2sdofmap[:, j].ravel()
        keep = np.isin(ix, self.freedofs) & np.isin(iy, self.freedofs)
        indices = np.stack([self.index_map[ix][keep], self.index_map[iy][keep]])
        return keep, indices

    @cached_property
    def index_map(self) -> NDArray[np.uint32]:
        indices = np.concatenate([self.freedofs, self.fixdofs])
        imap = np.zeros(len(indices), dtype=np.uint32)
        imap[indices] = np.arange(len(indices), dtype=np.uint32)
        return imap

    @cached_property
    def e2sdofmap(self) -> NDArray[np.uint32]:
        nx, ny, nz = self.out_shape
        z, x, y = np.unravel_index(np.arange(nx * ny * nz), (nz, nx, ny))
        idxs = self.dof_dim * (y + x * (ny + 1) + z * (nx + 1) * (ny + 1))
        return np.add(self.dofmap[None], idxs[:, None].astype(np.uint32))

    def global_mat(self, x):
        x = np.reshape(x, (-1, 1, 1)) * self.element[None]
        x = x.ravel()
        keep, indices = self.keep_indices
        return x[keep], indices

    def __call__(self, x, b):
        print("assemble matrix...", end=" ", flush=True)
        t0 = perf_counter()
        data, indices = self.global_mat(x)
        print(f"done: {perf_counter() - t0:.3f}s")

        print(f"solve using {self.solver}...", end=" ", flush=True)
        t0 = perf_counter()
        u_nz = solve_coo(data, indices, b.ravel()[self.freedofs], self.solver)
        print(f"done: {perf_counter() - t0:.3f}s")

        u = anp.concatenate([u_nz, np.zeros(len(self.fixdofs))])[self.index_map]

        dofmap = np.reshape(self.e2sdofmap.T, (-1, *self.out_shape))
        c = anp.einsum("izxy,ij,jzxy->xyz", u[dofmap], self.element, u[dofmap])

        return anp.sum(c)



#!/usr/bin/env python3

import autograd.numpy as np
import matplotlib.pyplot as plt
import nlopt
import pyvista as pv
from autograd import tensor_jacobian_product, value_and_grad

from tofea.fea3d import FEA3D_T
from tofea.topopt_helpers import simp_parametrization

max_its = 100
volfrac = 0.4
sigma = 1.0
shape = (80, 80, 80)
nx, ny, nz = shape
cmin, cmax = 1.0, 1e6

dofs = np.arange((nx + 1) * (ny + 1) * (nz + 1)).reshape(nz + 1, nx + 1, ny + 1)
fixed = np.zeros_like(dofs, dtype=bool)
load = np.zeros_like(dofs)

fixed[35:-35, 35:-35, 0] = 1
load[..., 5:] = 1

x0 = np.full(shape, volfrac)
fem = FEA3D_T(shape, dofs, fixed, load, z_chunks=10, solver="gpu")
parametrization = simp_parametrization(shape, sigma, cmin, cmax)

fig, ax = plt.subplots(2, 1)
im0 = ax[0].imshow(x0[..., x0.shape[-1] // 2].T, cmap="gray_r", vmin=0, vmax=1)
im1 = ax[1].imshow(
    parametrization(x0)[..., x0.shape[-1] // 2].T, cmap="gray_r", vmin=0, vmax=1
)
ax[0].axis("off")
ax[1].axis("off")
fig.tight_layout()
fig.show()


def volume_constraint(x, gd):
    v, g = value_and_grad(np.mean)(x)
    if gd.size > 0:
        gd[:] = g
    return v - volfrac


def nlopt_obj(x, gd):
    design = parametrization(x)
    c, dc = value_and_grad(fem)(design)

    if gd.size > 0:
        gd[:] = tensor_jacobian_product(parametrization)(x, dc).ravel()

    im0.set_array(np.reshape(x, design.shape)[..., design.shape[-1] // 2].T)
    im1.set_array(design[..., design.shape[-1] // 2].T)
    fig.canvas.draw()

    return c


opt = nlopt.opt(nlopt.LD_MMA, x0.size)
opt.add_inequality_constraint(volume_constraint)
opt.set_min_objective(nlopt_obj)
opt.set_lower_bounds(0)
opt.set_upper_bounds(1)
opt.set_maxeval(max_its)
design = opt.optimize(x0.ravel())

plt.show(block=True)

grid = pv.wrap(np.reshape(design, shape))
p = pv.Plotter()
p.add_volume(
    grid,
    opacity="linear",
    cmap="gray_r",
    show_scalar_bar=False,
    scalars="values",
)
p.set_background("w")
p.show()



import autograd.numpy as np
import scipy.ndimage
from autograd.extend import defvjp, primitive

gaussian_filter = primitive(scipy.ndimage.gaussian_filter)
defvjp(
    gaussian_filter,
    lambda ans, x, *args, **kwargs: lambda g: gaussian_filter(g, *args, **kwargs),
)


def sigmoid_projection(x, a, b=0.5):
    num = np.tanh(a * b) + np.tanh(a * (x - b))
    denom = np.tanh(a * b) + np.tanh(a * (1 - b))
    return num / denom


def sigmoid_parametrization(shape, sigma, alpha, beta=0.5, flat=False):
    def _parametrization(x):
        x = np.reshape(x, shape)
        x = gaussian_filter(x, sigma)
        x = sigmoid_projection(x, alpha, beta)
        return x.ravel() if flat else x

    return _parametrization


def simp_projection(x, vmin, vmax, penalty=3.0):
    return vmin + x**penalty * (vmax - vmin)


def simp_parametrization(shape, sigma, vmin, vmax, penalty=3.0, flat=False):
    def _parametrization(x):
        x = np.reshape(x, shape)
        x = gaussian_filter(x, sigma)
        x = simp_projection(x, vmin, vmax, penalty)
        return x.ravel() if flat else x

    return _parametrization


def gray_indicator(x):
    return np.mean(4 * x * (1 - x))



#!/usr/bin/env python3

import autograd.numpy as np
import matplotlib.pyplot as plt
import nlopt
from autograd import value_and_grad

from tofea.fea2d import FEA2D_T
from tofea.topopt_helpers import simp_parametrization

max_its = 100
volfrac = 0.4
sigma = 0.5
shape = (100, 100)
nelx, nely = shape
cmin, cmax = 1.0, 1e6

fixed = np.zeros((nelx + 1, nely + 1), dtype="?")
load = np.zeros_like(fixed)

fixed[40:60, -1] = 1
load[:, :-10] = 1

fem = FEA2D_T(fixed)
parametrization = simp_parametrization(shape, sigma, cmin, cmax)
x0 = np.full(shape, volfrac)


def objective(x):
    x = parametrization(x)
    x = fem(x, load)
    return x


def volume(x):
    x = parametrization(x)
    x = (x - cmin) / (cmax - cmin)
    return np.mean(x)


plt.ion()
fig, ax = plt.subplots(1, 1)
im = ax.imshow(parametrization(x0).T, cmap="gray_r", vmin=cmin, vmax=cmax)
fig.tight_layout()


def volume_constraint(x, gd):
    v, g = value_and_grad(volume)(x)
    if gd.size > 0:
        gd[:] = g
    return v - volfrac


def nlopt_obj(x, gd):
    c, dc = value_and_grad(objective)(x)

    if gd.size > 0:
        gd[:] = dc.ravel()

    im.set_data(parametrization(x).T)
    plt.pause(0.01)

    return c


opt = nlopt.opt(nlopt.LD_MMA, x0.size)
opt.add_inequality_constraint(volume_constraint, 1e-6)
opt.set_min_objective(nlopt_obj)
opt.set_lower_bounds(0)
opt.set_upper_bounds(1)
opt.set_maxeval(max_its)
opt.optimize(x0.ravel())

plt.show(block=True)



from autograd.extend import defjvp, defvjp, primitive
from scipy.sparse import coo_matrix

from tofea.solvers import get_solver

_default_solver = "scipy"


@primitive
def solve_coo(entries, indices, rhs, solver=_default_solver):
    a = coo_matrix((entries, indices)).tocsc()
    _solver = get_solver(solver)
    _solver.clear()
    _solver.factor(a)
    return _solver.solve(rhs)


def solve_coo_entries_jvp(g, x, entries, indices, b, solver=_default_solver):
    a = coo_matrix((g, indices)).tocsc()
    _solver = get_solver(solver)
    return _solver.solve(-(a @ x))


def solve_coo_b_jvp(g, x, entries, indices, b, solver=_default_solver):
    _solver = get_solver(solver)
    return _solver.solve(g)


defjvp(solve_coo, solve_coo_entries_jvp, None, solve_coo_b_jvp)


def solve_coo_entries_vjp(ans, entries, indices, b, solver=_default_solver):
    def vjp(g):
        _solver = get_solver(solver)
        x = _solver.solve(g, transpose=True)
        i, j = indices
        return -x[i] * ans[j]

    return vjp


def solve_coo_b_vjp(ans, entries, indices, b, solver=_default_solver):
    def vjp(g):
        _solver = get_solver(solver)
        return _solver.solve(g, transpose=True)

    return vjp


defvjp(solve_coo, solve_coo_entries_vjp, None, solve_coo_b_vjp)
